# Replaces the function RefineBins from ModIsom
RefineBins := function( o, bins, func, prnt )
    local i, j, k, res, inv, new;

    for i in [1..Length(bins)] do

        # report
        if prnt then
            Print("  start bin ",i," of ",Length(bins),"\n");
        fi;

        # compute values
        res := [];
        for j in [1..Length(bins[i])] do
            if IsInt(bins[i][j]) then
                res[j] := func(SmallGroup(o, bins[i][j]));
            else
                res[j] := func(bins[i][j]);
            fi;
        od;
        inv := Set(res);

        # split bin
        new := List(inv, x -> []);
        for j in [1..Length(res)] do
            k := Position(inv,res[j]);
            Add( new[k], bins[i][j] );
        od;

        # reset bin
        bins[i] := new;
    od;

    # return result
    if IsInt(bins[1][1]) then
        return SortedList(Concatenation(bins));
    else
        return Concatenation(bins);
    fi;
end;


#############################################################################
##
#F Group Theoretic Infos
## 
InstallMethod(MIPConjugacyClassInfo, [IsGroup], function(G)
    local p, e, c, s, o, exp, i, cov, r, cc, g, h;

    p := PrimePGroup(G);
    e := Elements(G);
    c := Orbits(G,e);

    # Roggenkamp
    s := [Sum(List(c, x -> RankPGroup(Stabilizer(G,x[1]))))];

    # the length
    Add( s, Length(c) );

    # p^n-th powers. Probably Kuelshammer
    repeat
        e := Set( List(e, x -> x^p) );
        o := Orbits(G, e);
        Add( s, Length(o) );
    until Length(e) = 1;

    # Parmenter+Polcino-Milies. S. Hertweck-Soriano Lemma 2.4. 
    #This can probably be written much more efficiently!
    exp := Log(Exponent(G), p);
    c := Filtered(c, x -> Size(x) > 1); # Filter out center to improve performance
    for i in [1..exp-1] do
        cov := [ ];
        r := 0;
        
        for cc in c do
            g := Representative(cc);
        
            if Order(g) < p^i then # Serves to improve performance. Does it?
                h := g^(p^i);
                if not h^G in cov then
                    if Size(g^G) = Size( h^G ) then
                        r := r + 1;
                        Add(cov, h^G);
                    fi;
                fi;
            fi;
        od;

        Add(s,r);
    od;

    return s;
end);


InstallMethod(MIPJenningsInfo, [IsGroup], function(G)
    local s, r, i, a;
    s := JenningsSeries(G);
    r := [];
    for i in [1..Length(s)-1] do
        a := [GroupInfo(s[i]/s[i+1])];
        if i <= Length(s)-2 then
            Add(a, GroupInfo(s[i]/s[i+2]));
        fi;
        if i <= (Length(s)-1)/2 then
            Add(a, GroupInfo(s[i]/s[2*i+1]));
        fi;
        Add(r, a);
    od;

    a := [ ];

    # Hertweck's G/M4(G) invariant
    if PrimePGroup(G) <> 2 then 
        # Not just use group info as it is weaker for orders 5^6 and 3^7 (ID_AVAILABLE fails for these orders)
        if Length(s) <= 4 and "IdGroup" in KnownAttributesOfObject(G) then
            Add(a, IdGroup(G)); 
        else
            Add(a, GroupInfo(G/s[4]));
        fi;
    # Not Hertweck, but can help for 2-groups of order 2^9
    else 
        # Not just use group info as it is weaker for orders 2^9 (ID_AVAILABLE fails for these orders)
        if Length(s) <= 3 and "IdGroup" in KnownAttributesOfObject(G) then
            Add(a, IdGroup(G)); 
        else
            Add(a, GroupInfo(G/s[3]));
        fi;
    fi;
    Add(r,a);
    return r;
end);


InstallMethod(MIPSandlingInfo, [IsGroup], function(G) 
    local s, p, U, r;
    s := LowerCentralSeries(G);
    p := PrimePGroup(G);
    r := [ ];
    U := Subgroup(G, Concatenation(Pcgs(s[3]), List(Pcgs(s[2]), x -> x^p)));
    if Size(U) = 1 and "IdGroup" in KnownAttributesOfObject(G) then
        Add(r,IdGroup(G));
    else
        Add(r, GroupInfo(G/U));
    fi;

    # Baginski
    if RankPGroup(G) <= 2 and Length(s) >= 4 then 
        U := Subgroup(G, Concatenation(Pcgs(s[4]), List(Pcgs(s[2]), x -> x^p)));
        if Size(U) = 1 and "IdGroup" in KnownAttributesOfObject(G) then
            Add(r,IdGroup(G));
        else
            Add(r, GroupInfo(G/U));
        fi;
    fi;
    return(r);
end);


###### Some new invariants. 
# Baginski 99 Corollary 7 and Theorem 9
BaginskiInfo := function(G) 
local D, N, act, F;
    D := DerivedSubgroup(G);
    F := FrattiniSubgroup(D);
 
    # if-condition only for technical reasons, output is the same
    if Size(D) = Size(F) then
        N := Centralizer(G, D);
    else
        act := AbelianSubfactorAction(G, D, FrattiniSubgroup(D));
        N := Kernel(act[1]);
    fi;
  
    if IsCyclic(G/N) then
        return [GroupInfo(G/FrattiniSubgroup(N)), GroupInfo(N/F)];
    else
        return false;
    fi;
end;


# Based on Theorem 6.11 from Sandling's 1984 survey
CenterDerivedInfo := function(G) 
local U;
    U := Intersection(Center(G), DerivedSubgroup(G));
   
    return [AbelianInvariants(U), AbelianInvariants(Center(G)/U)];
end;


# Hertweck-Soriano end of page 16
FrattiniInfo := function(G) 
local p;
  
    p := PrimePGroup(G);
    
    if IsElementaryAbelian(DerivedSubgroup(G)) then
        if Length(JenningsSeries(G)) <= (2*p) then
            return GroupInfo(FrattiniSubgroup(G));
        fi;
    fi;
  
    return false;
end;



# Based on Sandling survey Lemma 6.26 for G'
JenningsDerivedInfo := function(G) 
local s, r, a, i;
  
    s := JenningsSeries(DerivedSubgroup(G));
    r := [ ];

    for i in [1..Length(s)-1] do
        Add(r, GroupInfo(s[i]/s[i+1]));
    od;
  
    return r;
end;


# Based on Baginski-Caranti 1988 Proposition 1.2
BaginskiCarantiInfo := function(G) 
    return NilpotencyClassOfGroup(G/FrattiniSubgroup(DerivedSubgroup(G)));
end;


# Baginski-Konovalov 05
NilpotencyClassInfo := function(G) 
local p;
  
    p := PrimePGroup(G);

    # Note that all of these three are invariants anyway by other results
    if Exponent(G) = p or NilpotencyClassOfGroup(G) = 2 or IsCyclic(DerivedSubgroup(G)) then 
        return NilpotencyClassOfGroup(G);
    else
        return false;
    fi;
end;


DimensionTwoCohomology := function(G)
local p, M, L, C;

    p := PrimePGroup(G);
    L := List([1..Size(GeneratorsOfGroup(G))], x -> [[Z(p)^0]]);
    M := GModuleByMats(L, GF(p));
    C := TwoCohomology(G,M);

    return Dimension(Image(C.cohom));
end;


####################################
# Programs to test if group is covered by theoretical result. 
# Not including results for small groups
HasMaximalAbelianSubgroup := function(G)
local LM, U;

    LM := MaximalSubgroups(G);
  
    for U in LM do
        if IsAbelian(U) then
            return true;
        fi;
    od;
    
    return false;
end;
 
HasCyclicSubgroupIOfIndexP2 := function(G) 
local LM, U, LUM, V;

    LM := MaximalSubgroups(G);

    for U in LM do
        LUM := MaximalSubgroups(U);
    
        for V in LUM do
            if IsCyclic(V) then 
                return true;
            fi;
        od;
    od; 
    
    return false;
end;


# Some theoretical results follow from the other criteria. In particular metacyclic groups (Sandling 96) and (elem-ab.)-by-cyclic groups (Baginski 99) are covered
IsCoveredByTheory := function(G)
local p, n;

    p := PrimePGroup(G);
    n := Log(Size(G),p);

    # 2-groups of maximal class known. Baginski
    if p = 2 and NilpotencyClassOfGroup(G) = n-1 then 
        return true;
    fi;

    # Baginski-Caranti 88
    if p <> 2 and n <= p+1 and HasMaximalAbelianSubgroup(G) and NilpotencyClassOfGroup(G) = n-1 then 
        return true;
    fi;

    # Drensky 89
    if Size(G)/Size(Center(G)) = p^2 then 
        return true;
    fi;
 
    # Baginski-Konovalov 05. Rather slow, but can be helpful. Case p<>2 covered by other criteria by Lemma 1 (ArXiv-version) of BK05
    #if p = 2 and HasCyclicSubgroupIOfIndexP2(G) then 
        #return true;
    #fi;

    return false;
end;


#############################################################################
##
#O MIPBinsByGT( p, n, L )
##
##
InstallMethod( MIPBinsByGT, [IsPosInt, IsPosInt, IsList], function( p, n, L... )
local bins, cent, i;

    if not IsPrimeInt(p) then
        Error("<p> must be a prime integer");
    fi;

    if Length(L) = 1 then
        if not IsList(L[1]) or not (ForAll(Flat(L[1]), IsInt) or ForAll(Flat(L[1]), IsGroup)) then
            Error("<L> must be a list of lists of integers or a list of lists of groups");
        fi;

        bins := ShallowCopy(L[1]);
    else
        Error("number of arguments must be 2 or 3");
    fi;

    # refine by abelian invariants
    Info(InfoModIsomExt, 1, "refine by abelian invariants of group (Sehgal/Ward)");
    bins := RefineBins( p^n, bins, AbelianInvariants, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # refine by center
    Info(InfoModIsomExt, 1, "refine by abelian invariants of center (Sehgal/Ward)");
    cent := function(G) return AbelianInvariants(Center(G)); end;
    bins := RefineBins( p^n, bins, cent, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    ### NEW. 
    # refine by center and derived subgroup interplay
    Info(InfoModIsomExt, 1, "refine by abelian invariants of center with derived subgroup (Sandling)");
    bins := RefineBins( p^n, bins, CenterDerivedInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # refine by lower central series
    # Comment changed. This is really more accurate
    Info(InfoModIsomExt, 1, "refine by Sandling quotient (Sandling/Baginski)"); 
    bins := RefineBins( p^n, bins, MIPSandlingInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # CHANGED
    # refine by jennings series
    Info(InfoModIsomExt, 1, "refine by jennings series (Passi+Sehgal/Ritter+Sehgal/Hertweck)");
    bins := RefineBins( p^n, bins, MIPJenningsInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # NEW
    # refine by Jennings series of derived group
    Info(InfoModIsomExt, 1, "refine by jennings series of derived group (Sandling)");
    bins := RefineBins( p^n, bins, JenningsDerivedInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # NEW
    # refine by Baginski critiria
    Info(InfoModIsomExt, 1, "refine by Baginski criteria");
    bins := RefineBins( p^n, bins, BaginskiInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # NEW
    # refine by Baginski-Caranti critiria NilpotencyClass(G/Phi(G'))
    Info(InfoModIsomExt, 1, "refine by Baginski-Caranti invariant");
    bins := RefineBins( p^n, bins, BaginskiCarantiInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # NEW
    # refine by Frattini subgroup
    # From Hertweck-Soriano p.16
    Info(InfoModIsomExt, 1, "refine by Frattini subgroup for (el-ab)-by-ab (Hertweck+Soriano)");
    bins := RefineBins( p^n, bins, FrattiniInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # NEW
    # refine by Nilpotency class
    # From Baginski-Konovalov 05
    Info(InfoModIsomExt, 1, "refine by nilpotency class for some cases (Baginski+Konovalov)"); 
    bins := RefineBins( p^n, bins, NilpotencyClassInfo, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    #NEW
    # refine by dimension of cohomology  
    Info(InfoModIsomExt, 1, "refine by second cohomology group"); 
    bins := RefineBins( p^n, bins, DimensionTwoCohomology, false );
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # NEW
    # refine by known theoretical results
    Info(InfoModIsomExt, 1, "checking if covered by theory");
    for i in [1..Size(bins)] do
        if ForAll(bins[i], IsInt) then
            bins[i] := Filtered(bins[i], x -> not IsCoveredByTheory(SmallGroup(p^n, x)));
        else
            bins[i] := Filtered(bins[i], x -> not IsCoveredByTheory(x));
        fi; 
    od;
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # CHANGED
    # refine by conjugacy classes
    Info(InfoModIsomExt, 1, "refine by conjugacy classes (Parmenter+Polcino-Milies,Kuelshammer,Roggenkamp/Wursthorn)");
    if InfoLevel(InfoModIsomExt) >= 1 then
        bins := RefineBins( p^n, bins, MIPConjugacyClassInfo, true );
    else
        bins := RefineBins( p^n, bins, MIPConjugacyClassInfo, false );
    fi;
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups");
    if Length(bins)=0 then return bins; fi;

    # refine by subgroups
    Info(InfoModIsomExt, 1, "refine by elem-ab subgroups (Quillen)");
    if InfoLevel(InfoModIsomExt) >= 1 then
        bins := RefineBins( p^n, bins, SubgroupsInfo, true );
    else
        bins := RefineBins( p^n, bins, SubgroupsInfo, false );
    fi;
    bins := Filtered( bins, x -> Length(x)>1 );
    Info(InfoModIsomExt, 1, Length(bins)," bins with ",Length(Flat(bins))," groups \n");
    if Length(bins)=0 then return bins; fi;

    return bins;
end );


InstallOtherMethod( MIPBinsByGT, [IsPosInt, IsPosInt], function( p, n )

    if not IsPrimeInt(p) then
        Error("<p> must be a prime integer");
    fi;

    return MIPBinsByGT(p, n, [[1..NumberSmallGroups(p^n)]]);
end );


#############################################################################
##
#O JenningsBound( p, n, bin )
##
##
InstallMethod( JenningsBound, [IsPosInt, IsPosInt, IsList], function( p, n, bin )
local grps, js, i, ids, l, j;

    if not IsPrimeInt(p) then
        Error("<p> must be a prime integer");
    fi;

    # Check whether we have a list of integers or a list of groups
    if ForAll(bin, IsInt) then
        grps := List(bin, x -> SmallGroup(p^n, x));
    elif ForAll(bin, IsGroup) then
        grps := bin;
    else
        Error("<L> must be a list of integers or a list of groups");
    fi;

    js := List(grps, x -> JenningsSeries(x));
	
#    for i in Reversed([1..Length(js[1])]) do
#        ids := List(js, x -> IdSmallGroup(x[1]/x[i]));
#        if Length(Set(ids))=1 then
#            return i;
#        fi;
#    od;

    l := Minimum(List(js, Length));
	
    for i in [1..l] do
        if i = l then # this assumes that input groups are pairwise not isomorphic
          return l-1;
        fi;
        if ID_AVAILABLE(Size(js[1][1]/js[1][i])) <> fail then
            ids := List(js, x -> IdSmallGroup(x[1]/x[i]));
        else
            ids := [1];
            for j in [2..Size(grps)] do
                if IsomorphismGroups(js[1][1]/js[1][i], js[j][1]/js[j][i]) = fail then
                    ids := [1,2];
                    break;
                fi;
            od;
        fi;
        if Length(Set(ids))>1 then
            return i-1;
        fi;
    od;

end );


#############################################################################
##
#O JenningsBoundPairwise( p, n, bin )
##
##
InstallMethod( JenningsBoundPairwise, [IsPosInt, IsPosInt, IsList], function( p, n, bin )
local grps, js, i, ids, c, ret, idav;

    if not IsPrimeInt(p) then
        Error("<p> must be a prime integer");
    fi;

    if ID_AVAILABLE(p^n) <> fail then
        idav := true;
    else
        idav := false;	      
    fi;

    # Check whether we have a list of integers or a list of groups
    if ForAll(bin, IsInt) then
        grps := List(bin, x -> SmallGroup(p^n, x));
    elif ForAll(bin, IsGroup) then
        grps := bin;
    else
        Error("<L> must be a list of integers or a list of groups");
    fi;

    ret := [];

    for c in Combinations([1..Length(grps)], 2) do
        if idav then
            Add(ret, [IdSmallGroup(grps[c[1]]), IdSmallGroup(grps[c[2]]), JenningsBound(p, n, grps{c})]);
        elif "IdGroup" in KnownAttributesOfObject(grps[c[1]]) and "IdGroup" in KnownAttributesOfObject(grps[c[2]]) then
            Add(ret, [IdSmallGroup(grps[c[1]]), IdSmallGroup(grps[c[2]]), JenningsBound(p, n, grps{c})]);
        else
            Add(ret, [grps[c[1]], grps[c[2]], JenningsBound(p, n, grps{c})]);
        fi;
    od;

    return ret;
end );

#############################################################################
##Functions to check conjecture on Jennings bound from MM20
##
##
##
ConjecturedJenBound := function(p, jb)
if p = 2 then
  return 2*jb;
else
 return 2*jb - ((p-1)/2);
fi;
end;

InstallMethod( JenningsBoundConjecture, [IsPosInt, IsPosInt, IsRecord], function(p, n, tup)
local s, l, ids, jb, b, pair, jbp, bp, a, lp;
if IsString(tup.splits[1]) or IsString(tup.splits[1][1]) then
  Print("Data is given in format which does not allow the check. \n");
  return fail;
else
  s := tup.splits;
  l := Maximum(List(s, x -> x[1]));
  ids := Concatenation(List(s, x -> x[2]));
  jb := JenningsBound(p,n,ids);
  b := ConjecturedJenBound(p,jb);
  if l <= b then  
    return true;
  else
    for pair in Combinations(ids, 2) do
    jbp := JenningsBoundPairwise(p,n,pair)[1][3];
    bp := ConjecturedJenBound(p, jbp);
      if l > bp then
        a := MIPBinSplit(p,n,false,jbp,1,pair);
        lp := Maximum(List(a.splits, x -> x[1]));
        if lp > bp then
          return false;
        else 
          return true;
        fi;
      fi;
    od;
  fi;
fi;
end );

InstallMethod( JenningsBoundConjectureIsStrict, [IsPosInt, IsPosInt, IsRecord], function(p, n, tup)
local s, l, ids, jb, b;
if IsString(tup.splits[1]) or IsString(tup.splits[1][1]) then
  Print("Data is given in format which does not allow the check. \n");
  return fail;
else
  s := tup.splits;
  l := Maximum(List(s, x -> x[1]));
  ids := Concatenation(List(s, x -> x[2]));
  if Size(ids) > 3 then
    Print("This function is currently only implemented for pairs of groups \n");
    return fail;
  fi;
  jb := JenningsBound(p,n,ids);
  b := ConjecturedJenBound(p,jb);
  if l = b then  
    return true;
  else 
    return false;
  fi;
fi;
end );

