#############################################################################
##
#W    modisomext.gd                 The ModIsomExt package        Leo Margolis
#W                                                                Tobias Moede
##

DeclareInfoClass( "InfoModIsomExt" );

#############################################################################
##
#O  MIPBinSplit(<p>, <n>, <max>, <start>, <step>, <L>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="MIPBinSplit">
##  <ManSection>
##  <Oper Name="MIPBinSplit" Arg="p, n, max, start, step, L"/>
##
##  <Description>
##  Given a list <A>L</A> of small group library ids or a list of groups of order
##  <A>p</A>^<A>n</A> this functions checks isomorphism of the associated modular
##  group algebras using canonical forms for the quotients of the augmentation ideals
##  <M>I(kG)</M>. The parameter <A>max</A> is an integer or false that determines the 
##  maximal quotients <M>I(kG)/I(kG)^{<A>max</A>}</M> to be checked (if false is given as input,
##  then the quotients are enlarged until non-isomorphic quotients are found or 
##  eventually the full augmentation ideal will be checked). The parameter <A>start</A>
##  specifies which quotients <M>I(kG)/I(kG)^{<A>start</A>}</M> are precomputed. The
##  parameter <A>step</A> determines in which steps the quotients are enlarged
##  if necessary during the isomorphism check.
##  <P/>
##  The output is a record containing three entries: <C>bins</C> contains all the groups, for
##  which the non-isomorphism of the associated modular group algebras could not be determined; 
##  <C>splits</C> contains all the groups, for which the associated group algebras were determined
##  to be non-isomorphic (and the first non-isomorphic quotient); <C>time</C> contains the
##  time used for the computation (in milliseconds).
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPBinSplit", [IsPosInt, IsPosInt, IsObject, IsPosInt, IsPosInt, IsList] );


#############################################################################
##
#O  TableOfRadQuotient(<kG>, <n>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="TableOfRadQuotient">
##  <ManSection>
##  <Oper Name="TableOfRadQuotient" Arg="kG, n"/>
##
##  <Description>
##  Given a modular group algebra <A>kG</A> the function computes the class-<A>n</A>
##  quotient of the augmentation ideal <M>I(kG)</M>, i.e. <M>I(kG)/I(kG)^{<A>n</A>+1}</M>. The output is a nilpotent 
##  table for this quotient. Note that in addition to the standard entries of
##  a nilpotent table it contains further entries for computational reasons.
##  This allows do determine the class-<A>n</A> quotient of the augmentation ideal
##  without computing the full augmentation ideal using <C>NilpotentTableOfRad</C>
##  as provided by the ModIsom package. 
##  <P/>
##  The components <A>dim</A>,  <A>fld</A>, <A>rnk</A>, <A>tab</A>, <A>wgs</A>, <A>wds</A>
##  remain unchanged from the ModIsom package. The additional components are
##   <A>commwords</A>, <A>powwords</A> and <A>pre</A>. These new components
##  contain additional information on precisely which basis of <M>I(kG)/I(kG)^{<A>n</A>+1}</M> is used
##  and what the result of multiplying basis elements is. We explain how users
##  can understand how the basis looks and how they can multiply two elements in the algebra. 
##  <P/>
##  The dimension of <M>I(kG)/I(kG)^{<A>n</A>+1}</M> is recorded in <A>T.dim</A>.
##  The basis of <M>I(kG)/I(kG)^{<A>n</A>+1}</M> is found as in the theory of Jennings, cf.
##  <Cite Key="MaMo20"/>. The elements of <M>G</M> chosen to provide the basis
##  of subsequent quotients of dimension subgroups are recorded in <A>T.pre.jen.pcgs</A>.
##  Let us call these elements <M>g_1,\ldots,g_m</M>. Note that <M>|G| = p^m</M>.
##  If now <M>l</M> is an integer smaller than
##  <A>T.dim+1</A>, then the <M>l</M>-th elements of the basis of  <M>I(kG)/I(kG)^{<A>n</A>+1}</M> 
##  is <M>(g_1-1)^{e_1} \cdot \ldots (g_m-1)^{e_m}</M> where <A> [e_1,...,e_m] = T.pre.exps[l]</A>.
##  The weight of this element is recorded in <A>T.wgs[l]</A> and also <A>T.pre.weights[l]</A>.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "TableOfRadQuotient", [IsGroupRing, IsPosInt] );


#############################################################################
##
#O  MIPBinsByGT(<p>, <n>, <L>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="MIPBinsByGT">
##  <ManSection>
##  <Oper Name="MIPBinsByGT" Arg="p, n, [L]"/>
##
##  <Description>
##  Given a list <A>L</A> of small group library ids or a list of groups of order
##  <A>p</A>^<A>n</A> the function uses group theoretical invariants to split the
##  groups into bins. If <A>L</A> is not given, then all groups of order p^n are
##  considered.
## 
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPBinsByGT", [IsPosInt, IsPosInt, IsList] );


#############################################################################
##
#O  JenningsBound(<p>, <n>, <L>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="JenningsBound">
##  <ManSection>
##  <Oper Name="JenningsBound" Arg="p, n, L"/>
##
##  <Description>
##  Given a list <A>L</A> of small group library ids or a list of groups of order
##  <A>p</A>^<A>n</A> the function computes an integer <M>b</M> such that the
##  quotients of the associated augmentation ideals are guaranteed to be isomorphic
##  up to class <M>b-1</M>. 
##  <P/>
##  More precisely <M>b</M> is the biggest integer such that for any <M>G, H \in L</M>
##  one has <M>G/D_b(G) \cong H/D_b(H)</M>.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "JenningsBound", [IsPosInt, IsPosInt, IsList] );


#############################################################################
##
#O  JenningsBoundPairwise(<p>, <n>, <L>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="JenningsBoundPairwise">
##  <ManSection>
##  <Oper Name="JenningsBoundPairwise" Arg="p, n, L"/>
##
##  <Description>
##  Given a list <A>L</A> of small group library ids or a list of groups of order
##  <A>p</A>^<A>n</A> the function computes for all pairs <M>(G,H)</M> of groups in 
##  the list an integer <M>b</M> such that the quotients of the associated augmentation 
##  ideals are guaranteed to be isomorphic up to class <M>b-1</M>. 
##  <P/>
##  More precisely the return is a list of triples. The first two entries of each triple
##  are two groups <M>G</M> and <M>H</M>, or their ids if they are available, and the
##  last entry contains <A>JenningsBound(p,n,[G,H])</A>.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "JenningsBoundPairwise", [IsPosInt, IsPosInt, IsList] );

#############################################################################
##
#O  JenningsBoundConjecture(<p>, <n>, <L>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="JenningsBoundConjecture">
##  <ManSection>
##  <Oper Name="JenningsBoundConjecture" Arg="p, n, L"/>
##
##  <Description>
##  Given a list <A>L</A> in the format returned by <C>MIPBinSplit</C> for
##  some groups of order <M>p^n</M> this function checks if the groups violate
##  the bound conjectured in Question 2.7 <Cite Key="MaMo20"/>
##  on the maximal quotients of the corresponding augmentation ideals which
##  need to be checked to decide MIP. 
##  <P/>
##  For elements of <C>MIPResults(p,n)</C> which are solved by theoretical results,
##  or which remain open, it returns <C>fail</C>.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "JenningsBoundConjecture", [IsPosInt, IsPosInt, IsRecord] );

#############################################################################
##
#O  JenningsBoundConjectureIsStrict(<p>, <n>, <L>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="JenningsBoundConjectureIsStrict">
##  <ManSection>
##  <Oper Name="JenningsBoundConjectureIsStrict" Arg="p, n, L"/>
##
##  <Description>
##  Given a list <A>L</A> in the format returned by <C>MIPBinSplit</C> for
##  some groups of order <M>p^n</M> this function checks if the groups attain
##  the bound conjectured in Question 2.7 <Cite Key="MaMo20"/>
##  on the maximal quotients of the corresponding augmentation ideals which
##  need to be checked to decide MIP. This function currently only works if
##  <A>L</A> contains the information for a pair of groups.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "JenningsBoundConjectureIsStrict", [IsPosInt, IsPosInt, IsRecord] );

#############################################################################
##
#O  MIPConjugacyClassInfo(<G>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="MIPConjugacyClassInfo">
##  <ManSection>
##  <Oper Name="MIPConjugacyClassInfo" Arg="G"/>
##
##  <Description>
##  For a given <M>p</M>-group <A>G</A> this function returns a list <A>L</A> containing known
##  group-theoretical invariants associated to the conjugacy classes of <A>G</A>. 
##  The first entry of <A>L</A> is the so-called Roggenkamp parameter
##  <M>\sum_{g^G} \log_p(|C_G(g)/\Phi(C_G(g))| )</M> where the sum runs over conjugacy classes of <A>G</A>.
##  The next entries contain the number of conjugacy classe which are <M>p^\ell</M>-th powers,
##  for <M>0 \leq \ell \leq log_p(exp(G)) </M> (Kuelshammer). Note that for <M>\ell=0</M> this is just the number
##  of conjugacy classes in <A>G</A>.
##  Finally, the following entries contain the number of conjugacy classes of <M>p^\ell</M>-th powers
##  which are not central and have the same order as a class which powers to them
##  where <M>0 \leq \ell \leq log_p(exp(G))-1</M> (Parmenter-Polcino Milies). 
##
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPConjugacyClassInfo", [IsGroup] );

#############################################################################
##
#O  MIPJenningsInfo(<G>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="MIPJenningsInfo">
##  <ManSection>
##  <Oper Name="MIPJenningsInfo" Arg="G"/>
##
##  <Description>
##  For a given <A>p</A>-group <A>G</A> this function returns a list <A>L</A> containing
##  information on quotients of the Jennings-Zassenhaus series <M>D_i(G)</M> of <A>G</A>.
##  Starting with <A>i=1</A> for increasing <A>i</A> it contatins the
##  <A>GroupInfo</A> for <M>D_i(G)/D_{i+1}(G)</M>, <M>D_i(G)/D_{i+2}(G)</M>,
##  and <M>D_i(G)/D_{2i+1}(G)</M> when these are defined.
##  The last entry describes <M>G/D_3(G)</M> if <A>p=2</A> and <M>G/D_4(G)</M>
##  if <A>p>2</A>. If <M>D_3(G)=1</M> or <M>D_4(G)=1</M>, respectively, and 
##  <A>IdGroup</A> is a known attribute of <A>G</A>, it is <A>IdGroup(G)</A>. 
##  Otherwise it contains the <A>GroupInfo</A> of <M>G/D_3(G)=1</M> or <M>G/D_4(G)=1</M>
##  respectively.
##
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPJenningsInfo", [IsGroup] );


#############################################################################
##
#O  MIPSandlingInfo(<G>)  . . . . . . . . . . 
##
##  <#GAPDoc Label="MIPSandlingInfo">
##  <ManSection>
##  <Oper Name="MIPSandlingInfo" Arg="G"/>
##
##  <Description>
##  For a given <A>p</A>-group <A>G</A> this function returns a list <A>L</A> containing
##  information on the Sandling quotient <M>G/\gamma_2(G)^p\gamma_3(G)</M>.
##  The first entry describes <M>Q = G/\gamma_2(G)^p\gamma_3(G)</M> in the following way:
##  If <M>\gamma_2(G)^p\gamma_3(G) = 1</M> and <A>IdGroup</A> is a known attribute of <A>G</A>,
##  it is <A>IdGroup(G)</A>. Otherwise it contains the <A>GroupInfo</A> of 
##  <M>G/\gamma_2(G)^p\gamma_3(G)</M> (Sandling).
##  Moreover, if <A>G</A> is generated by at most two elements and the length of 
##  the Jennings-Zasenhaus series of <A>G</A> is at least four, it contains a second entry
##  describing <M>G/\gamma_2(G)^p\gamma_4(G)</M> in a similar way (Baginski/Margolis-Moede).
##
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPSandlingInfo", [IsGroup] );


#############################################################################
##
#O  MIPBins(<p>, <n>)  . . . . . . . . . . . . . . . .
##
##  <#GAPDoc Label="MIPBins">
##  <ManSection>
##  <Oper Name="MIPBins" Arg="p, n"/>
##
##  <Description>
##  Given a prime <A>p</A> and a positive integer <A>n</A> the function returns
##  the bins resulting from using group theoretical invariants. Precomputed bins
##  are only available for orders <M>2^6</M>, <M>2^7</M>, <M>2^8</M>, <M>2^9</M>, <M>3^6</M>,
##  <M>3^7</M> and <M>5^6</M>.
##  Note that in the case <M>2^9</M> only 2- and 3-generated groups of this order
##  are considered.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPBins", [IsPosInt, IsPosInt] );


#############################################################################
##
#O  MIPResults(<p>, <n>)  . . . . . . . . . . . . . . . .
##
##  <#GAPDoc Label="MIPResults">
##  <ManSection>
##  <Oper Name="MIPResults" Arg="p, n"/>
##
##  <Description>
##  Given a prime <A>p</A> and a positive integer <A>n</A> the function returns
##  the results of verifying MIP. This is only available for orders <M>2^6</M>, 
##  <M>2^7</M>, <M>2^8</M>, <M>2^9</M>, <M>3^6</M>, <M>3^7</M> and <M>5^6</M>.
##  Note that in the case <M>2^9</M> only 2- and 3-generated groups of this order
##  are considered.
##  
##  </Description>
##  </ManSection>
##  <#/GAPDoc>
DeclareOperation( "MIPResults", [IsPosInt, IsPosInt] );

#E  modisomext.gd  . . . . . . . . . . . . . . . . . . . . . . . . ends here
